# -*- coding: utf-8 -*-
import warnings


class DisclaimerRecord(object):
    """
    Represents a disclaimer entry with the label and it's message.

    Args:
        title (dict of unicode): The multilingual disclaimer message label
        content (dict of unicode): The multilingual disclaimer message.
    """
    def __init__(self, title, content):
        """
        Attributes:
            title (dict of unicode): The multilingual disclaimer message label
            content (dict of unicode): The multilingual disclaimer message.
        """
        if not isinstance(title, dict):
            warnings.warn('Type of "title" should be "dict"')
        if not isinstance(content, dict):
            warnings.warn('Type of "content" should be "dict"')

        self.title = title
        self.content = content



# -*- coding: utf-8 -*-
from pyramid_oereb import route_prefix
from pyramid_oereb.core.views.webservice import PlrWebservice, Symbol, Logo, Sld
from pyramid_oereb.contrib.stats.decorators import log_response


def includeme(config):  # pragma: no cover
    """
    Create all necessary routes and views for the `pyramid_oereb` application.

    Args:
        config (pyramid.config.Configurator): The application's configurator instance.
    """

    # Service for logo images
    config.add_route('{0}/image/logo'.format(route_prefix),
                     '/image/logo/{logo}/{language}.{extension}')
    config.add_view(Logo, attr='get_image', route_name='{0}/image/logo'.format(route_prefix),
                    request_method='GET',
                    decorator=log_response)

    # Service for symbol images
    config.add_route('{0}/image/symbol'.format(route_prefix),
                     '/image/symbol/{theme_code}/legend_entry.{extension}')

    config.add_view(Symbol, attr='get_image', route_name='{0}/image/symbol'.format(route_prefix),
                    request_method='GET',
                    decorator=log_response)

    # Service for sld creation on egrid input
    config.add_route('{0}/sld'.format(route_prefix), '/sld')
    config.add_view(Sld, attr='get_sld', route_name='{0}/sld'.format(route_prefix),
                    request_method='GET',
                    decorator=log_response)

    # Get versions
    config.add_route('{0}/versions'.format(route_prefix), '/versions/{format}')
    config.add_view(
        PlrWebservice,
        attr='get_versions',
        route_name='{0}/versions'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )
    config.add_route('{0}/versions/'.format(route_prefix), '/versions/{format}/')
    config.add_view(
        PlrWebservice,
        attr='get_versions',
        route_name='{0}/versions/'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )

    # Get capabilities
    config.add_route('{0}/capabilities'.format(route_prefix), '/capabilities/{format}')
    config.add_view(
        PlrWebservice,
        attr='get_capabilities',
        route_name='{0}/capabilities'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )
    config.add_route('{0}/capabilities/'.format(route_prefix), '/capabilities/{format}/')
    config.add_view(
        PlrWebservice,
        attr='get_capabilities',
        route_name='{0}/capabilities/'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )

    # Get egrid
    config.add_route('{0}/getegrid'.format(route_prefix),
                     '/getegrid/{format}')
    config.add_view(
        PlrWebservice,
        attr='get_egrid',
        route_name='{0}/getegrid'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )
    config.add_route('{0}/getegrid/'.format(route_prefix),
                     '/getegrid/{format}/')
    config.add_view(
        PlrWebservice,
        attr='get_egrid',
        route_name='{0}/getegrid/'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )

    # Get extract by id
    config.add_route('{0}/extract'.format(route_prefix),
                     '/extract/{format}')
    config.add_view(
        PlrWebservice,
        attr='get_extract_by_id',
        route_name='{0}/extract'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )
    config.add_route('{0}/extract/'.format(route_prefix),
                     '/extract/{format}/')
    config.add_view(
        PlrWebservice,
        attr='get_extract_by_id',
        route_name='{0}/extract/'.format(route_prefix),
        request_method='GET',
        decorator=log_response
    )

    # Commit config
    config.commit()



# -*- coding: utf-8 -*-
import os

import logging
import pyaml_env
from pyramid.config import ConfigurationError

from pyramid_oereb.core.readers.availability import AvailabilityReader
from pyramid_oereb.core.readers.disclaimer import DisclaimerReader
from pyramid_oereb.core.readers.glossary import GlossaryReader
from pyramid_oereb.core.readers.municipality import MunicipalityReader
from pyramid_oereb.core.records.office import OfficeRecord
from pyramid_oereb.core.records.document_types import DocumentTypeRecord
from pyramid_oereb.core.records.law_status import LawStatusRecord
from pyramid_oereb.core.records.real_estate_type import RealEstateTypeRecord
from pyramid_oereb.core.readers.theme import ThemeReader
from pyramid_oereb.core.readers.theme_document import ThemeDocumentReader
from pyramid_oereb.core.records.logo import LogoRecord
from pyramid_oereb.core.readers.logo import LogoReader
from pyramid_oereb.core.readers.law_status import LawStatusReader
from pyramid_oereb.core.readers.real_estate_type import RealEstateTypeReader
from pyramid_oereb.core.readers.document_types import DocumentTypeReader
from pyramid_oereb.core.readers.document import DocumentReader
from pyramid_oereb.core.readers.office import OfficeReader
from pyramid_oereb.core.readers.general_information import GeneralInformationReader
from pyramid_oereb.core.readers.map_layering import MapLayeringReader
from sqlalchemy.exc import ProgrammingError

log = logging.getLogger(__name__)


class Config(object):
    """
    A central point where we can access to the application configuration.
    Init it with a config file (Config.init(configfile, configsection))
    Then use it anywhere with Config.get('my_config_variable').
    """

    _config = None
    themes = None
    logos = None
    document_types = None
    documents = None
    general_information = None
    law_status = None
    real_estate_types = None
    map_layering = None
    theme_document = None
    offices = None
    availabilities = None
    glossaries = None
    disclaimers = None
    municipalities = None

    @staticmethod
    def init(configfile, configsection, c2ctemplate_style=False, init_data=False):
        """
        Loads configuration from yaml file and provides methods for generating webservice output.

        Args:
            config_file (str): The configuration yaml file.
            configsection (str): The section within the yaml file.
            c2ctemplate_style (bool): If set to true, c2c.template library will be used
                to load config file (Default: False).
            init_data (bool): Indicates wheather or not data should be initialized (Default: False).
        """

        assert Config._config is None

        Config._config = _parse(configfile, configsection, c2ctemplate_style)
        if init_data:
            Config.init_law_status()
            Config.init_document_types()
            Config.init_offices()
            Config.init_themes()
            Config.init_theme_document()
            Config.init_documents()
            Config.init_general_information()
            Config.init_real_estate_types()
            Config.init_map_layering()
            Config.init_logos()
            Config.assemble_relation_themes_documents()
            Config.init_availabilities()
            Config.init_glossaries()
            Config.init_disclaimers()
            Config.init_municipalities()

    @staticmethod
    def get_config():
        """
        Returns the current configuration.

        Returns:
            dict: The current config or None.
        """

        return Config._config

    @staticmethod
    def update_settings(settings):
        """
        Updates the current pyramid configuration settings.

        Args:
            settings (dict): Dictionary with the current pyramid settings.
        """
        settings.update(Config._config)

    @staticmethod
    def init_themes():
        """
        Initializes all themes configured in config file.

        Raises:
            ProgrammingError
        """
        try:
            Config.themes = Config._read_themes()
        # When initializing the database (create_tables), the table 'theme' does not exist yet
        except ProgrammingError:
            Config.themes = None

    @staticmethod
    def init_theme_document():
        """
        Initializes all theme documents object configured in config file.

        Raises:
            ProgrammingError
        """
        try:
            Config.theme_document = Config._read_theme_document()
        except ProgrammingError:
            Config.theme_document = None

    @staticmethod
    def init_general_information():
        """
        Initializes general information object configured in config file.

        Raises:
            ProgrammingError
        """
        try:
            Config.general_information = Config._read_general_information()
        except ProgrammingError:
            Config.general_information = None

    @staticmethod
    def init_law_status():
        """
        Initializes law status object configured in config file.

        Raises:
            ProgrammingError
        """
        try:
            Config.law_status = Config._read_law_status()
        except ProgrammingError:
            Config.law_status = None

    @staticmethod
    def init_documents():
        """
        Initializes documents object configured in config file.

        Raises:
            ProgrammingError
        """
        try:
            Config.documents = Config._read_documents()
        except ProgrammingError:
            Config.documents = None

    @staticmethod
    def init_offices():
        """
        Initializes offices object configured in config file.

        Raises:
            ProgrammingError
        """
        try:
            Config.offices = Config._read_offices()
        except ProgrammingError:
            Config.offices = None

    @staticmethod
    def init_availabilities():
        """
        Initializes availabilities from the configured source.
        """
        Config.availabilities = Config._read_availabilities()

    @staticmethod
    def init_glossaries():
        Config.glossaries = Config._read_glossaries()

    @staticmethod
    def init_disclaimers():
        Config.disclaimers = Config._read_disclaimers()

    @staticmethod
    def init_municipalities():
        Config.municipalities = Config._read_municipalities()

    @staticmethod
    def assemble_relation_themes_documents():
        """
        Loads theme document records based on Config.themes, Config.theme_document and Config.documents.
        """

        if Config.theme_document is not None:
            for theme in Config.themes:
                theme_documents = []
                for theme_document in Config.theme_document:
                    if theme_document.theme_id == theme.identifier:
                        for document in Config.documents:
                            if theme_document.document_id == document.identifier:
                                document_copy = document.copy()
                                document_copy.article_numbers = theme_document.article_numbers
                                theme_documents.append(document_copy)
                if len(theme_documents) > 0:
                    theme.document_records = theme_documents
        else:
            log.info('No global documents related to themes were provided!')

    @staticmethod
    def _read_themes():
        """
        Reads the theme settings from config and initiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.theme.ThemeRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        theme_config = Config.get_theme_config()
        if theme_config is None:
            raise ConfigurationError("Missing configuration for themes")
        theme_reader = ThemeReader(
            theme_config.get('source').get('class'),
            **Config.get_theme_config().get('source').get('params')
        )
        return theme_reader.read()

    @staticmethod
    def _read_theme_document():
        """
        Reads the theme document settings from config and initiates the relevant reader.

        Returns:
        list of pyramid_oereb.core.records.theme_document.ThemeDocumentRecord:
            The list of found records. Since these are not filtered by any criteria the list simply
            contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        theme_document_config = Config.get_theme_document_config()
        if theme_document_config is None:
            raise ConfigurationError("Missing configuration for theme document")
        theme_document_reader = ThemeDocumentReader(
            theme_document_config.get('source').get('class'),
            **Config.get_theme_document_config().get('source').get('params')
        )
        return theme_document_reader.read()

    @staticmethod
    def _read_general_information():
        """
        Reads settings of general infomation from config an intiates the relevant reader.

        Returns:
        list of pyramid_oereb.core.records.general_information.GeneralInformationRecord:
            The list of found records. Since these are not filtered by any criteria the list simply
            contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        info_config = Config.get_info_config()
        if info_config is None:
            raise ConfigurationError("Missing configuration for general information")
        info_reader = GeneralInformationReader(
            info_config.get('source').get('class'),
            **Config.get_info_config().get('source').get('params')
        )
        return info_reader.read()

    @staticmethod
    def _read_law_status():
        """
        Reads settings of law status from config an intiates the relevant reader.

        Returns:
        list of pyramid_oereb.core.records.lawstatus.LawStatusRecord:
            The list of found records. Since these are not filtered by any criteria the list simply
            contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        law_status_config = Config.get_law_status_config()
        if law_status_config is None:
            raise ConfigurationError("Missing configuration for law status source config")
        law_status_reader = LawStatusReader(
            law_status_config.get('source').get('class'),
            **law_status_config.get('source').get('params')
        )
        return law_status_reader.read()

    @staticmethod
    def _read_documents():
        """
        Reads settings of documents from config an intiates the relevant reader.

        Returns:
        list of pyramid_oereb.core.records.documents.DocumentRecord:
            The list of found records. Since these are not filtered by any criteria the list simply
            contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        document_config = Config.get_document_config()
        if document_config is None:
            raise ConfigurationError("Missing configuration for document source config")
        document_reader = DocumentReader(
            document_config.get('source').get('class'),
            **document_config.get('source').get('params')
        )
        return document_reader.read(Config.offices)

    @staticmethod
    def _read_offices():
        """
        Reads settings of offices from config and instantiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.office.OfficeRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        office_config = Config.get_office_config()
        if office_config is None:
            raise ConfigurationError("Missing configuration for office source config")
        office_reader = OfficeReader(
            office_config.get('source').get('class'),
            **office_config.get('source').get('params')
        )
        return office_reader.read()

    @staticmethod
    def _read_availabilities():
        """
        Reads settings of availabilities from configured source and instantiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.office.OfficeRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        availability_config = Config.get_availability_config()
        if availability_config is None:
            raise ConfigurationError("Missing configuration for availability source config")
        availability_reader = AvailabilityReader(
            availability_config.get('source').get('class'),
            **availability_config.get('source').get('params')
        )
        return availability_reader.read()

    @staticmethod
    def _read_glossaries():
        """
        Reads settings of glossaries from configured source and instantiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.glossary.GlossaryRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """
        glossary_config = Config.get_glossary_config()
        if glossary_config is None:
            raise ConfigurationError("Missing configuration for glossary source config")
        glossary_reader = GlossaryReader(
            glossary_config.get('source').get('class'),
            **glossary_config.get('source').get('params')
        )
        return glossary_reader.read()

    @staticmethod
    def _read_disclaimers():
        """
        Reads settings of disclaimers from configured source and instantiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.disclaimer.DisclaimerRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """
        disclaimer_config = Config.get_disclaimer_config()
        if disclaimer_config is None:
            raise ConfigurationError("Missing configuration for disclaimer source config")
        disclaimer_reader = DisclaimerReader(
            disclaimer_config.get('source').get('class'),
            **disclaimer_config.get('source').get('params')
        )
        return disclaimer_reader.read()

    @staticmethod
    def _read_municipalities():
        """
        Reads settings of disclaimers from configured source and instantiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.municipality.MunicipalityRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """
        municipality_config = Config.get_municipality_config()
        if municipality_config is None:
            raise ConfigurationError("Missing configuration for municipality source config")
        municipality_reader = MunicipalityReader(
            municipality_config.get('source').get('class'),
            **municipality_config.get('source').get('params')
        )
        return municipality_reader.read()

    @staticmethod
    def get_srid():
        """
        Returns the srid.

        Returns:
            int: The srid if it was set.
        """
        assert Config._config is not None
        return Config._config.get('srid')

    @staticmethod
    def get_general_information():
        """
        Returns the general information.

        Returns:
            list of pyramid_oereb.core.records.theme.GeneralInformationRecord: The available general
            information entries.
        """

        assert Config._config is not None
        return Config.general_information

    @staticmethod
    def get_law_status_codes():
        """
        Returns a list of available law status codes.

        Returns:
            list of unicode: The available law status codes.
        """

        assert Config._config is not None
        return [law_status.code for law_status in Config.law_status]

    @staticmethod
    def get_themes():
        """
        Returns a list of available themes.

        Returns:
            list of pyramid_oereb.core.records.theme.ThemeRecord: The available themes.
        """

        assert Config._config is not None
        return Config.themes

    @staticmethod
    def get_theme_by_code_sub_code(code, sub_code=None):
        """
        Returns the theme or sub-theme which matches the code.

        Args:
            code (str): The theme's code.
            sub_code (str): The theme's sub code (Default: None).

        Returns:
            pyramid_oereb.core.records.theme.ThemeRecord or None: The theme with the specified
            code.

        Raises:
            ConfigurationError
        """

        if Config.themes is None:
            raise ConfigurationError("Themes have not been initialized")
        theme_result = [theme for theme in Config.themes if theme.sub_code == sub_code and theme.code == code]
        if len(theme_result) > 0:
            return theme_result[0]
        else:
            raise ConfigurationError(
                f"Theme {code} with sub-code {sub_code} not found in the application configuration"
            )

    @staticmethod
    def init_logos():
        """
        Initializes logos object list configured in config file.

        Raises:
            ProgrammingError
        """

        try:
            Config.logos = Config._read_logos()
        # When initializing the database (create_tables), the table 'logo' does not exist yet
        except ProgrammingError:
            Config.logos = None

    @staticmethod
    def _read_logos():
        """
        Reads settings of logos from config an intiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.logo.LogoRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        logo_config = Config.get_logo_config()
        if logo_config is None:
            raise ConfigurationError("Missing configuration for logos")
        logo_reader = LogoReader(
            logo_config.get('source').get('class'),
            **Config.get_logo_config().get('source').get('params')
        )
        return logo_reader.read()

    @staticmethod
    def get_logos():
        """
        Returns all the logos and municipalities arms of coats.

        Returns:
            list of pyramid_oereb.core.records.logo.LogoRecord: All the logo entries needed to
            generate an plr data-extract (plr-logo, confederation, canton, municipality).
        """

        assert Config._config is not None
        if len(Config.logos) < 1:
            raise ConfigurationError("At least one entry for the plr-logo is required")
        return Config.logos

    @staticmethod
    def get_logo_by_code(code):
        """
        Returns the image for a logo called by its code.

        Args:
            code (str): The identifier for the logo.

        Returns:
            pyramid_oereb.core.records.logo.LogoRecord or None: The logo image
            for the specified code.

        Raises:
            ConfigurationError
        """

        if Config.logos is None:
            raise ConfigurationError("The logo images have not been initialized")
        for logo in Config.logos:
            if isinstance(logo, LogoRecord):
                if logo.code == code:
                    return logo
            else:
                raise ConfigurationError("The logo has not the expected format")
        raise ConfigurationError(f"Logo for code: {code} not found in the application configuration")

    @staticmethod
    def get_conferderation_logo():
        """
        Returns the image of confederation logo.

        Returns:
            pyramid_oereb.core.records.logo.LogoRecord or None: The logo image.

        """

        return Config.get_logo_by_code(Config.get_logo_lookup_confederation())

    @staticmethod
    def get_canton_logo():
        """
        Returns the image of canton logo.

        Returns:
            pyramid_oereb.core.records.logo.LogoRecord or None: The logo image.
        """

        return Config.get_logo_by_code(Config.get_logo_lookup_canton())

    @staticmethod
    def get_municipality_logo(fosnr):
        """
        Returns the image of municipality logo by fosnr.

        Args:
            fosnr (int): id of municipality.

        Returns:
            pyramid_oereb.core.records.logo.LogoRecord or None: The logo image.
        """

        return Config.get_logo_by_code('{}.{}'.format(
            Config.get_logo_lookup_confederation(),
            str(fosnr))
        )

    @staticmethod
    def get_oereb_logo():
        """
        Returns the image of oereb logo.

        Returns:
            pyramid_oereb.core.records.logo.LogoRecord or None: The logo image.
        """

        return Config.get_logo_by_code(Config.get_logo_lookup_oereb())

    @staticmethod
    def get_logo_lookups():
        """
        Reads logo config settings.

        Returns:
            Dict of logo lookup settings.

        Raises:
            ConfigurationError
        """

        logo_lookups = Config.get('logo_lookups')
        if logo_lookups:
            return logo_lookups
        else:
            raise ConfigurationError('Configuration for "logo_lookups" not found.')

    @staticmethod
    def get_logo_lookup(key):
        """
        Reads logo config settings by its key.

        Args:
            key (str): the key of the logo to look for in config file.

        Returns:
            str: Code of logo provided in config for the specified key.

        Raises:
            ConfigurationError
        """

        code = Config.get_logo_lookups()[key]
        if code:
            return code
        else:
            raise ConfigurationError('Configuration for lookup "{}" not found.'.format(key))

    @staticmethod
    def get_logo_lookup_oereb():
        """
        Returns the oereb logo code specified in config file.

        Returns:
            str: Code of oereb logo provided in config file.
        """

        return Config.get_logo_lookup('oereb')

    @staticmethod
    def get_logo_lookup_canton():
        """
        Returns the canton logo code specified in config file.

        Returns:
            str: Code of canton logo provided in config file.
        """

        return Config.get_logo_lookup('canton')

    @staticmethod
    def get_logo_lookup_confederation():
        """
        Returns the confederation logo code specified in config file.

        Returns:
            str: Code of confedertaion logo provided in config.
        """

        return Config.get_logo_lookup('confederation')

    @staticmethod
    def init_document_types():
        """
        Initiates all document types configured in the config file.

        Raises:
            ProgrammingError
        """

        try:
            Config.document_types = Config._read_document_types()
        # When initializing the database (create_tables), the table 'document_types' does not exist yet
        except ProgrammingError:
            Config.document_types = None

    @staticmethod
    def _read_document_types():
        """
        Reads settings of document types from config an intiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.document_types.DocumentTypeRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        document_types_config = Config.get_document_types_config()
        if document_types_config is None:
            raise ConfigurationError("Missing configuration for document types")
        document_types_reader = DocumentTypeReader(
            document_types_config.get('source').get('class'),
            **Config.get_document_types_config().get('source').get('params')
        )
        return document_types_reader.read()

    @staticmethod
    def get_document_types():
        """
        Returns a list of available document_types.

        Returns:
            list of pyramid_oereb.core.records.document_types.DocumentTypeRecord: The available
            document types.
        """

        assert Config._config is not None
        return Config.document_types

    @staticmethod
    def init_real_estate_types():
        """
        Initializes all real estate types configured in the config file.

        Raises:
            ProgrammingError
        """

        try:
            Config.real_estate_types = Config._read_real_estate_types()
        # When initializing the database (create_tables), the table 'real_estate_type' does not exist yet
        except ProgrammingError:
            Config.real_estate_types = None

    @staticmethod
    def _read_real_estate_types():
        """
        Reads settings of real estate types from config an intiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.realestatetype.RealEstateTypeRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        real_estate_types_config = Config.get('real_estate_type')
        if real_estate_types_config is None:
            raise ConfigurationError("Missing configuration for real estate type source config")
        real_estate_type_reader = RealEstateTypeReader(
            real_estate_types_config.get('source').get('class'),
            **real_estate_types_config.get('source').get('params'))

        return real_estate_type_reader.read()

    @staticmethod
    def get_real_estate_types():
        """
        Returns a list of available real_estate_types.

        Returns:
            list of pyramid_oereb.core.records.real_estate_type.RealEstateTypeRecord: The available
            real estate types.
        """

        assert Config._config is not None
        return Config.real_estate_types

    @staticmethod
    def init_map_layering():
        """
        Initializes map layering as configured in the config file.

        Raises:
            ProgrammingError
        """

        try:
            Config.map_layering = Config._read_map_layering()
        # When initializing the database (create_tables), the table 'map_layering' does not exist yet
        except ProgrammingError:
            Config.map_layering = None

    @staticmethod
    def _read_map_layering():
        """
        Reads settings of map layering from config an intiates the relevant reader.

        Returns:
            list of pyramid_oereb.core.records.map_layering.MapLayeringRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.

        Raises:
            ConfigurationError
        """

        map_layering_config = Config.get('map_layering')
        if map_layering_config is None:
            raise ConfigurationError("Missing configuration for map layering source config")
        map_layering_reader = MapLayeringReader(
            map_layering_config.get('source').get('class'),
            **map_layering_config.get('source').get('params'))

        return map_layering_reader.read()

    @staticmethod
    def get_map_layering():
        """
        Returns a list of available map layerings.

        Returns:
            list of pyramid_oereb.core.records.map_layering.MapLayeringRecord: The available
            map layering.
        """

        assert Config._config is not None
        return Config.map_layering

    @staticmethod
    def get_index_and_opacity_of_view_service(reference_wms):
        """
        Returns the index and the opacity of the view service.

        Args:
            reference_wms (dict): reference_wms.

        Returns:
            pyramid_oereb.standard.models.main.MapLayering.layer_index or 1 (default value),
            pyramid_oereb.standard.models.main.MapLayering.layer_opacity or 0.75 (default value)
        """

        default_index = 1
        default_opacity = 0.75
        if Config.map_layering is None:
            raise ConfigurationError("Map layering has not been initialized")
        if len(Config.map_layering) == 0:
            return default_index, default_opacity
        for map_layering_result in Config.map_layering:
            for lang1 in list(reference_wms.keys()):
                for lang2 in list(map_layering_result.keys()):
                    if reference_wms[lang1] == map_layering_result[lang2]:
                        return map_layering_result.layer_index, map_layering_result.layer_opacity
        return default_index, default_opacity

    @staticmethod
    def get_document_types_lookups(theme_code):
        """
        Returns the document types config specified in config file for given theme.

        Args:
            theme_code (str): theme code to get lookups for from config file.

        Returns:
            dict: config settings of document types

        Raises:
            ConfigurationError
        """

        lookups = Config.get_theme_config_by_code(theme_code).get('document_types_lookup')
        if lookups is None:
            raise ConfigurationError(
                '"document_types_lookup" must be defined in configuration for theme {}!'.format(theme_code)
            )
        return lookups

    @staticmethod
    def get_document_type_lookup_by_theme_code_key_code(theme_code, key, code):
        """
        Returns the document types config specified in config file for given theme and key.

        Args:
            theme_code (str): theme code to get lookups for from config file.
            key (str): key of lookup pair
            code (str): value of lookup pair

        Returns:
            dict: config settings of document types

        Raises:
            ConfigurationError
        """

        lookups = Config.get_document_types_lookups(theme_code)
        for lookup in lookups:
            if lookup[key] == code:
                return lookup
        raise ConfigurationError(
            'Document type lookup for theme {} with key "{}" and code "{}" is not '
            'defined in configuration!'.format(theme_code, key, code)
        )

    @staticmethod
    def get_document_type_lookup_by_data_code(theme_code, data_code):
        """
        Returns the document types config specified in config file for given theme data_code.

        Args:
            theme_code (str): theme code to get lookups for from config file.
            data_code (str): key of lookup pair

        Returns:
            dict: config settings of document types
        """

        return Config.get_document_type_lookup_by_theme_code_key_code(
            theme_code,
            'data_code',
            data_code
        )

    @staticmethod
    def get_document_type_by_data_code(theme_code, data_code):
        """
        Returns the document type record for given theme data_code.

        Args:
            theme_code (str): theme code to get lookups for from config file.
            data_code (str): key of lookup pair.

        Returns:
            pyramid_oereb.core.records.document_types.DocumentTypeRecord or None:
                document types record of theme and data_code.
        """

        lookup = Config.get_document_type_lookup_by_data_code(theme_code, data_code)
        record = Config.get_document_type_by_code(lookup['transfer_code'])
        log.debug(
            'Translating code {} => code {} of {}'.format(
                lookup['data_code'], lookup['extract_code'], record.title
            )
        )
        translated_record = DocumentTypeRecord(lookup['extract_code'], record.title)
        return translated_record

    @staticmethod
    def get_main_document_types_lookups():
        """
        Returns the document types config specified in config file in app_schema.

        Returns:
            dict: config settings of document types.

        Raises:
            ConfigurationError
        """

        lookups = Config._config.get('app_schema').get('document_types_lookup')
        if lookups is None:
            raise ConfigurationError(
                '"document_types_lookup" must be defined in "app_schema"!'
            )
        return lookups

    @staticmethod
    def get_main_document_type_lookup_by_key_code(key, code):
        """
        Returns the document types config specified in config file in app_schema.

        Args:
            key (str): key of lookup pair.
            code (str): value of lookup pair.

        Returns:
            dict: config settings of document types.

        Raises:
            ConfigurationError
        """

        lookups = Config.get_main_document_types_lookups()
        for lookup in lookups:
            if lookup[key] == code:
                return lookup
        raise ConfigurationError(
            'Document type lookup with key "{}" and code "{}" is not '
            'defined in configuration!'.format(key, code)
        )

    @staticmethod
    def get_main_document_type_lookup_by_data_code(data_code):
        """
        Returns the document type record for given data_code.

        Args:
            data_code (str): key of lookup pair.

        Returns:
            dict: config settings of document types.
        """

        return Config.get_main_document_type_lookup_by_key_code(
            'data_code',
            data_code
        )

    @staticmethod
    def get_main_document_type_by_data_code(data_code):
        """
        Returns the document type record for given data_code.

        Args:
            data_code (str): key of lookup pair.

        Returns:
            pyramid_oereb.core.records.document_types.DocumentTypeRecord or None:
                document types record of data_code.
        """

        lookup = Config.get_main_document_type_lookup_by_data_code(data_code)
        record = Config.get_document_type_by_code(lookup['transfer_code'])
        log.debug(
            'Translating code {} => code {} of {}'.format(
                lookup['data_code'], lookup['extract_code'], record.title
            )
        )
        translated_record = DocumentTypeRecord(lookup['extract_code'], record.title)
        return translated_record

    @staticmethod
    def get_document_type_by_code(code):
        """
        Returns the label for the document type by code.

        Args:
            code (str): The document type code.

        Returns:
            pyramid_oereb.core.records.document_types.DocumentTypeRecord or None: The document
            type for the specified code.

        Raises:
            ConfigurationError
        """

        if Config.document_types is None:
            raise ConfigurationError("The document types have not been initialized")

        for document_type in Config.document_types:
            if document_type.code == code:
                return document_type
        raise ConfigurationError(f"Document type {code} not found in the application configuration")

    @staticmethod
    def get_theme_thresholds(code):
        """
        Returns the limits for the geometries of the theme with the specified code.

        Args:
            code (str): The theme's code.

        Returns:
            dict or None: The geometric tolerances for this theme.
        """
        assert Config._config is not None

        plrs = Config._config.get('plrs')
        if plrs and isinstance(plrs, list):
            for theme in plrs:
                if theme.get('code') == code:
                    return theme.get('plr_thresholds')
        return None

    @staticmethod
    def get_all_federal():
        """
        Returns a list of all federal topic codes.

        Returns:
            list (str): All federal topic codes.
        """

        assert Config._config is not None
        federal = []
        plrs = Config.get('plrs')
        if plrs and isinstance(plrs, list):
            for plr in plrs:
                if plr.get('federal'):
                    federal.append(plr.get('code'))
        return federal

    @staticmethod
    def get_crs():
        """
        Returns a unicode string of configured crs.

        Returns:
            unicode (str): The available crs.
        """

        assert Config._config is not None
        srid = Config._config.get('srid')
        return u'epsg:{}'.format(srid)

    @staticmethod
    def get_language():
        """
        Returns a list of available languages.

        Returns:
            list (str): The available languages.
        """

        assert Config._config is not None

        result = []
        language = Config._config.get('language')
        if language and isinstance(language, list):
            result.extend(language)
        return result

    @staticmethod
    def get_flavour():
        """
        Returns a list of available flavours.

        Returns:
            list (str): The available flavours.
        """

        assert Config._config is not None

        result = []
        flavour = Config._config.get('flavour')
        if flavour and isinstance(flavour, list):
            result.extend(flavour)
        return result

    @staticmethod
    def get_geometry_types():
        """
        Returns a list of available geometry_types.

        Returns:
            list (str): The available geometry types.
        """

        assert Config._config is not None

        result = []
        geometry_types = Config._config.get('geometry_types')
        if geometry_types and isinstance(geometry_types, list):
            result.extend(geometry_types)
        return result

    @staticmethod
    def get_real_estate_config():
        """
        Returns a dictionary of the configured real estate settings.

        Returns:
            dict: The configured real estate settings.
        """

        assert Config._config is not None

        return Config._config.get('real_estate')

    @staticmethod
    def get_real_estate_type_config():
        """
        Returns a dictionary of the configured real estate type settings.

        Returns:
            dict: The configured real estate type settings.
        """

        assert Config._config is not None

        return Config._config.get('real_estate_type')

    @staticmethod
    def get_plan_for_land_register_main_page_config():
        """
        Returns dictionary of plan for land register for the main page from the config file.

        Returns:
            dict: The config of the main page for the plan for land register.
        """

        assert Config._config is not None
        return Config._config.get('real_estate', {}).get('plan_for_land_register_main_page')

    @staticmethod
    def get_plan_for_land_register_config():
        """
        Returns dictionary of plan for land register from the config file.

        Returns:
            dict: The config for the plan for land register.
        """

        assert Config._config is not None
        return Config._config.get('real_estate', {}).get('plan_for_land_register')

    @staticmethod
    def get_address_config():
        """
        Returns a dictionary of the configured address settings.

        Returns:
            dict: The configured address settings.
        """

        assert Config._config is not None

        return Config._config.get('address')

    @staticmethod
    def get_theme_config():
        """
        Returns a dictionary of the configured theme settings.

        Returns:
            dict: The configured theme settings.
        """

        assert Config._config is not None

        return Config._config.get('theme')

    @staticmethod
    def get_theme_document_config():
        """
        Returns a dictionary of the configured theme document settings.

        Returns:
            dict: The configured theme document settings.
        """

        assert Config._config is not None

        return Config._config.get('theme_document')

    @staticmethod
    def get_document_types_config():
        """
        Returns a dictionary of the configured document type values settings.

        Returns:
            dict: The configured document types settings.
        """

        assert Config._config is not None

        return Config._config.get('document_types')

    @staticmethod
    def get_glossary_config():
        """
        Returns a dictionary of the configured glossary settings.

        Returns:
            dict: The configured glossary settings.
        """

        assert Config._config is not None
        return Config._config.get('glossary')

    @staticmethod
    def get_disclaimer_config():
        """
        Returns a dictionary of the configured disclaimer settings.

        Returns:
            dict: The configured disclaimer settings.
        """

        assert Config._config is not None

        return Config._config.get('disclaimer')

    @staticmethod
    def get_info_config():
        """
        Returns a dictionary of the configured general settings.

        Returns:
            dict: The configured general information settings.
        """

        assert Config._config is not None

        return Config._config.get('general_information')

    @staticmethod
    def get_logo_config():
        """
        Returns a dictionary of the configured file path's to the logos.

        Returns:
            dict: The configured paths to the logos wrapped in a dictionary.
        """

        assert Config._config is not None

        return Config._config.get('logos')

    @staticmethod
    def get_law_status_config():
        """
        Returns a dictionary of the configured law status sources.

        Returns:
            dict: The configured general law status sources.
        """

        assert Config._config is not None

        return Config._config.get('law_status_labels')

    @staticmethod
    def get_document_config():
        """
        Returns a dictionary of the configured document sources.

        Returns:
            dict: The configured documents sources.
        """

        assert Config._config is not None

        return Config._config.get('documents')

    @staticmethod
    def get_office_config():
        """
        Returns a dictionary of the configured office sources.

        Returns:
            dict: The configured office sources.
        """

        assert Config._config is not None

        return Config._config.get('offices')

    @staticmethod
    def get_municipality_config():
        """
        Returns a dictionary of the configured municipality settings.

        Returns:
            dict: The configured municipality settings.
        """

        assert Config._config is not None

        return Config._config.get('municipality')

    @staticmethod
    def get_extract_config():
        """
        Returns a dictionary of the configured extract settings.

        Returns:
            dict: The configured extract settings.
        """

        assert Config._config is not None

        return Config._config.get('extract')

    @staticmethod
    def get_availability_config():
        """
        Returns a dictionary of the configured availability settings.

        Returns:
            dict: The configured availability settings.
        """
        assert Config._config is not None

        return Config._config.get('availability')

    @staticmethod
    def get_plr_cadastre_authority():
        """
        Returns an office record for the configured PLR cadastre authority.

        Returns:
            pyramid_oereb.lib.records.office.OfficeRecord: The configured PLR cadastre
            authority.
        """

        assert Config._config is not None

        cfg = Config._config.get('plr_cadastre_authority')
        return OfficeRecord(
            cfg.get('name'),
            uid=cfg.get('uid'),
            office_at_web=cfg.get('office_at_web'),
            line1=cfg.get('line1'),
            line2=cfg.get('line2'),
            street=cfg.get('street'),
            number=cfg.get('number'),
            postal_code=cfg.get('postal_code'),
            city=cfg.get('city')
        )

    @staticmethod
    def get_oereblex_config():
        """
        Returns a dictionary of the configured OEREBlex settings.

        Returns:
            dict: The configured OEREBlex settings.
        """

        assert Config._config is not None

        return Config._config.get('oereblex')

    @staticmethod
    def get(key, default=None):
        """
        Returns the specified configuration value.

        Args:
            key (str): Configuration parameter name.
            default (*): Default value if the specified parameter is not defined. Defaults to
                None.

        Returns:
            *: The specified configuration or default value
        """

        assert Config._config is not None

        return Config._config.get(key, default)

    @staticmethod
    def get_object_path(path, default=None, required=None):
        """
        Returns the configuration object at a specified path.

        example:
        get_object_path('app.print', {'dpi': 300}, ['map_size'])

        Args:
            path (str): Dot separated path of the wanted object.
            default (dict): Default dictionary values of the object. Defaults to {}.
            required (list): The list of required sub values in the object. Defaults to [].

        Returns:
            *: The specified configuration object.
        """

        return Config._get_object_path(
            [], Config._config, path.split('.'),
            default if default is not None else {},
            required if required is not None else [])

    @staticmethod
    def _get_object_path(current_path, current_object, path, default, required):
        """
        Iterates through the given path list to return a valid dictionary of of the wanted object.

        Args:
            current_path (list): Path of current object.
            current_object (dict): Dictionary of current object.
            path (list): List of path entries to the desired object.
            default (dict): Default dictionary values of the object. Defaults to {}.
            required (list): The list of required sub values in the object. Defaults to [].

        Returns:
            *: The specified configuration object.

        Raises:
            ConfigurationError
        """

        if len(path) == 0:
            result = dict(default)
            result.update(current_object)
            for key in required:
                if key not in result:
                    raise ConfigurationError('Missing configuration value for {}.{}.'.format(
                        '.'.join(current_path), key))
            return result

        k = path[0]
        if k not in current_object:
            raise ConfigurationError('Missing configuration object for {}.{}.'.format(
                '.'.join(current_path), k))

        current_path.append(k)

        if type(current_object[k]) != dict:
            raise ConfigurationError('The configuration {} is not an object.'.format(
                '.'.join(current_path)))

        return Config._get_object_path(current_path, current_object[k], path[1:], default, required)

    @staticmethod
    def get_law_status_lookups(theme_code):
        """
        Returns the law status config specified in config file.

        Args:
            theme_code (str): Theme code to retrieve lookups for.

        Returns:
            dict: config settings of law status.

        Raises:
            ConfigurationError
        """

        lookups = Config.get_theme_config_by_code(theme_code).get('law_status_lookup')
        if lookups is None:
            raise ConfigurationError(
                '"document_types_lookup" must be defined in configuration for theme {}!'.format(theme_code)
            )
        return lookups

    @staticmethod
    def get_law_status_lookup_by_theme_code_key_code(theme_code, key, code):
        """
        Returns the law status config specified in config file for the specified theme and key.

        Args:
            theme_code (str): Theme code to retrieve lookups for.
            key (str): Key to retrieve lookups for.
            code (str): Value of the desired law status code.

        Returns:
            dict: config settings of law status.

        Raises:
            ConfigurationError
        """

        lookups = Config.get_law_status_lookups(theme_code)
        for lookup in lookups:
            if lookup[key] == code:
                return lookup
        raise ConfigurationError(
            'Law status lookup for theme {} with key "{}" and code "{}" is not '
            'defined in configuration!'.format(theme_code, key, code)
        )

    @staticmethod
    def get_law_status_lookup_by_data_code(theme_code, data_code):
        """
        Returns the law status config specified in config file.

        Args:
            theme_code (str): Theme code to retrieve lookups for.
            data_code (str): Value of the desired law status code.

        Returns:
            dict: config settings of law status.
        """
        return Config.get_law_status_lookup_by_theme_code_key_code(
            theme_code,
            'data_code',
            data_code
        )

    @staticmethod
    def get_law_status_by_data_code(theme_code, data_code):
        """
        Returns the law status record.

        Args:
            theme_code (str): Theme code to retrieve lookups for.
            data_code (str): Value of the desired law status code.

        Returns:
            pyramid_oereb.core.records.lawstatus.LawStatusRecord: The law status record.

        """
        lookup = Config.get_law_status_lookup_by_data_code(theme_code, data_code)
        record = Config.get_law_status_by_code(lookup['transfer_code'])
        log.debug(
            'Translating code {} => code {} of {}'.format(
                lookup['data_code'], lookup['extract_code'], record.title
            )
        )
        translated_record = LawStatusRecord(lookup['extract_code'], record.title)
        return translated_record

    @staticmethod
    def get_main_law_status_lookups():
        """
        Returns the main law status config specified in config file.

        Returns:
            dict: config settings of main law status.

        Raises:
            ConfigurationError
        """
        lookups = Config._config.get('app_schema').get('law_status_lookup')
        if lookups is None:
            raise ConfigurationError(
                '"law_status_lookup" must be defined in "app_schema"!'
            )
        return lookups

    @staticmethod
    def get_main_law_status_lookup_by_key_code(key, code):
        """
        Returns the main law status config specified in config file by key and code.

        Args:
            key (str): Key to retrieve lookups for.
            code (str): Value of the desired main law status code.

        Returns:
            dict: config settings of main law status.

        Raises:
            ConfigurationError
        """

        lookups = Config.get_main_law_status_lookups()
        for lookup in lookups:
            if lookup[key] == code:
                return lookup
        raise ConfigurationError(
            'Document type lookup with key "{}" and code "{}" is not'
            'defined in configuration!'.format(key, code)
        )

    @staticmethod
    def get_main_law_status_lookup_by_data_code(data_code):
        """
        Returns the main law status config specified in config file by data_code.

        Args:
            data_code (str): Value of the desired main law status code.

        Returns:
            dict: config settings of main law status.
        """

        return Config.get_main_law_status_lookup_by_key_code(
            'data_code',
            data_code
        )

    @staticmethod
    def get_main_law_status_by_data_code(data_code):
        """
        Returns the main law status record by data_code.

        Args:
            data_code (str): Value of the desired main law status code.

        Returns:
            pyramid_oereb.core.records.lawstatus.LawStatusRecord: The main law status record.
        """

        lookup = Config.get_main_law_status_lookup_by_data_code(data_code)
        record = Config.get_law_status_by_code(lookup['transfer_code'])
        log.debug(
            'Translating code {} => code {} of {}'.format(
                lookup['data_code'], lookup['extract_code'], record.title
            )
        )
        translated_record = LawStatusRecord(lookup['extract_code'], record.title)
        return translated_record

    @staticmethod
    def get_law_status_by_code(law_status_code):
        """
         Returns the law status record by law status code.

        Args:
            law_status_code (str): The law status code.

        Returns:
            pyramid_oereb.lib.records.law_status.LawStatusRecord: The found record.
        """

        if Config.law_status is None:
            raise ConfigurationError("The law status have not been initialized")

        for record in Config.law_status:
            if record.code == law_status_code:
                return record
        raise ConfigurationError(f"Law status {law_status_code} not found in the application configuration")

    @staticmethod
    def get_theme_config_by_code(theme_code):
        """
        Obtaining the theme configuration from the config file.

        Args:
            theme_code (str): The theme code.

        Returns:
            dict: theme settings from config.
        """

        assert Config._config is not None
        themes = Config._config.get('plrs')
        if themes and isinstance(themes, list):
            for theme in themes:
                if theme.get('code').lower() == theme_code.lower():
                    return theme
        return None

    @staticmethod
    def get_layer_config(theme_code):
        """
        Obtaining the layer configuration of a theme from config.

        Args:
            theme_code (str): The theme code.

        Returns:
            list: Layer index (int) and layer opacity (float).
        """
        assert Config._config is not None
        themes = Config._config.get('plrs')
        if themes and isinstance(themes, list):
            for theme in themes:
                if theme.get('code') == theme_code:
                    view_service = theme.get('view_service')
                    if view_service and isinstance(view_service, dict):
                        layer_index = view_service.get('layer_index')
                        layer_opacity = view_service.get('layer_opacity')
                        if layer_opacity is None:
                            raise ConfigurationError(
                                'For {} the "layer_opacity" was not found!'.format(theme_code)
                            )
                        if layer_index is None:
                            raise ConfigurationError(
                                'For {} the "layer_index" was not found!'.format(theme_code)
                            )
                        return layer_index, layer_opacity
        return None, None

    @staticmethod
    def get_real_estate_type_lookups():
        """
        Returns the real estate type config specified in config file.

        Returns:
            dict: config settings of main law status.

        Raises:
            ConfigurationError
        """

        lookups = Config.get_real_estate_type_config().get('lookup')
        if lookups is None:
            raise ConfigurationError(
                '"lookup" must be defined in configuration for real_estate_type!'
            )
        return lookups

    @staticmethod
    def get_real_estate_type_lookup_by_key_code(key, code):
        """
        Returns the real estate type lookup specified in config file by key and code.

        Args:
            key (str): Key to retrieve lookups for.
            code (str): Value of the desired real restate type code.

        Returns:
            dict: config settings of real restate type.

        Raises:
            ConfigurationError
        """

        lookups = Config.get_real_estate_type_lookups()
        for lookup in lookups:
            if lookup[key] == code:
                return lookup
        raise ConfigurationError(
            'Real estate type lookup with key "{}" and code "{}" is not '
            'defined in configuration!'.format(key, code)
        )

    @staticmethod
    def get_real_estate_type_lookup_by_data_code(data_code):
        """
        Returns the real estate type lookup specified in config file by data_code.

        Args:
            data_code (str): Value of the desired real estate type code.

        Returns:
            dict: config settings of real restate type.
        """

        return Config.get_real_estate_type_lookup_by_key_code('data_code', data_code)

    @staticmethod
    def get_real_estate_type_by_data_code(data_code):
        """
        Returns the real estate type file by data_code.

        Args:
            data_code (str): Value of the desired real estate type code.

        Returns:
            pyramid_oereb.core.records.realestatetype.RealEstateTypeRecord: The real estate type record.
        """
        lookup = Config.get_real_estate_type_lookup_by_data_code(data_code)
        record = Config.get_real_estate_type_by_code(lookup['transfer_code'])
        log.debug(
            'Translating code {} => code {} of {}'.format(
                data_code, lookup['extract_code'], record.title
            )
        )
        translated_record = RealEstateTypeRecord(lookup['extract_code'], record.title)
        return translated_record

    @staticmethod
    def get_real_estate_type_by_code(code):
        """
         Returns the record of the real estate types of the given code.

        Args:
            code (str): The real estate type code.

        Returns:
            pyramid_oereb.lib.records.real_estate_type.RealEstateTypeRecord: The found record.

        Raises:
            ConfigurationError
        """
        if Config.real_estate_types is None:
            raise ConfigurationError("The real estate types have not been initialized")

        for record in Config.real_estate_types:
            if record.code == code:
                return record
        raise ConfigurationError(f"Real estate type with {code} not found in the application configuration")

    @staticmethod
    def extract_module_function(dotted_function_path):
        """
        Returns a dictionary of the given function and its module path.

        Args:
            dotted_function_path (str): Path to functin in dot syntax.

        Returns:
            dict: given function and its module path.
        """

        elements = dotted_function_path.split('.')
        function_name = elements[-1]
        module_path = '.'.join(elements[:-1])
        return {
            'module_path': module_path,
            'function_name': function_name
        }

    @staticmethod
    def get_bbox(geometry):
        """
        Return a bbox adapted for the map size specified in the print configuration
         and based on the geometry and a buffer (margin to add between the geometry
         and the border of the map).

        Args:
            geometry (shapely.geometry.base.BaseGeometry): The geometry to calculate the bbox for.

        Returns:
            list: The bbox (meters) for the print.
        """
        print_conf = Config.get_object_path('print', required=['basic_map_size', 'buffer'])
        print_buffer = print_conf['buffer']
        map_size = print_conf['basic_map_size']
        map_width = float(map_size[0])
        map_height = float(map_size[1])

        if print_buffer * 2 >= min(map_width, map_height):
            error_msg_txt = 'Your print buffer ({}px)'.format(print_buffer)
            error_msg_txt += ' applied on each side of the feature exceed the smaller'
            error_msg_txt += ' side of your map {}px.'.format(min(map_width, map_height))
            raise ConfigurationError(error_msg_txt)

        geom_bounds = geometry.bounds
        geom_width = float(geom_bounds[2] - geom_bounds[0])
        geom_height = float(geom_bounds[3] - geom_bounds[1])

        geom_ratio = geom_width / geom_height
        map_ratio = map_width / map_height

        # If the format of the map is naturally adapted to the format of the geometry
        is_format_adapted = geom_ratio > map_ratio

        if is_format_adapted:
            # Part (percent) of the margin compared to the map width.
            margin_in_percent = 2 * print_buffer / map_width
            # Size of the margin in geom units.
            geom_margin = geom_width * margin_in_percent
            # Geom width with the margins (right and left).
            adapted_geom_width = geom_width + (2 * geom_margin)
            # Adapted geom height to the map height
            adapted_geom_height = (adapted_geom_width / map_width) * map_height
        else:
            # Part (percent) of the margin compared to the map height.
            margin_in_percent = 2 * print_buffer / map_height
            # Size of the margin in geom units.
            geom_margin = geom_height * margin_in_percent
            # Geom height with the buffer (top and bottom).
            adapted_geom_height = geom_height + (2 * geom_margin)
            # Adapted geom width to the map width
            adapted_geom_width = (adapted_geom_height / map_height) * map_width

        height_to_add = (adapted_geom_height - geom_height) / 2
        width_to_add = (adapted_geom_width - geom_width) / 2

        return [
            geom_bounds[0] - width_to_add,
            geom_bounds[1] - height_to_add,
            geom_bounds[2] + width_to_add,
            geom_bounds[3] + height_to_add,
        ]

    @staticmethod
    def get_map_size(extract_format):
        """
        Calculates the image size for the map if extract_format is "pdf" or returns configured size.

        Args:
            extract_format (str): The format which is requested. If it is pdf some special
                handling takes place to provide a HIDPI image.
        Returns:
            list of int: The size wrapped in a list of the form [width, height] in pixels.
        """
        print_conf = Config.get_object_path('print', required=['basic_map_size',
                                                               'pdf_dpi', 'pdf_map_size_millimeters'])
        if extract_format != 'pdf':
            return [int(print_conf['basic_map_size'][0]), int(print_conf['basic_map_size'][1])]
        else:
            pixel_size = print_conf['pdf_dpi'] / 25.4
            map_size_mm = print_conf['pdf_map_size_millimeters']
            return [int(pixel_size * map_size_mm[0]), int(pixel_size * map_size_mm[1])]

    @staticmethod
    def get_db_vars_from_env():
        """
        return the DB connection parameters below as a dict from the environment
        """
        DB_VARS = ["PGHOST", "PGPORT", "PGUSER", "PGPASSWORD", "PGDATABASE"]
        return {db_var: os.environ[db_var] for db_var in DB_VARS if db_var in os.environ}

    @staticmethod
    def availability_by_theme_code_municipality_fosnr(theme_code, fosnr):
        """
        Looks for configured availabilities. If a match was found it returns the matched
        availability status (bool).

        If none was found with the given theme code, fosnr it implicates the status is published.

        Args:
            theme_code (str): The theme code to look for.
            fosnr (int): The fosnr (aka ID_BFS) to look for.

        Returns:
             bool: True if the combination of municipality and theme code is available, false otherwise.
        """
        if Config.availabilities is None:
            raise ConfigurationError("The availabilities have not been initialized")
        for availability in Config.availabilities:
            if int(fosnr) == int(availability.fosnr) and theme_code == availability.theme_code:
                return availability.available
        return True

    @staticmethod
    def municipality_by_fosnr(fosnr):
        """
        Loops through all configured municipalities read from configured source to find a match by
        fosnr identifier.

        Args:
            fosnr (int): The key/fosnr which is used to find matching municipality.

        Returns:
            pyramid_oereb.lib.records.municipality.MunicipalityRecord: The found municipality
        Raises:
            ConfigurationError: If no match was found

        """
        for municipality in Config.municipalities:
            if municipality.fosnr == fosnr:
                return municipality
        raise ConfigurationError(
            'No municipalitiy with fosnr {} could be found in the configured municipalities ({}).'.format(
                fosnr,
                Config.municipalities
            )
        )


def _parse(cfg_file, cfg_section, c2ctemplate_style=False):
    """
    Parses the defined YAML file and returns the defined section as dictionary.

    Args:
        cfg_file (str): The YAML file to be parsed.
        cfg_section (str): The section to be returned.
        c2ctemplate_style (bool): If set to true, c2c.template library
            will be used to load config file (Default: False).

    Returns:
        dict: The parsed section as dictionary.

    Raises:
        ConfigurationError

    Throws:
        IOError
    """

    if cfg_file is None:
        raise ConfigurationError(
            'Missing configuration parameter "pyramid_oereb.cfg.file" or '
            '"pyramid_oereb.cfg.c2ctemplate.file".'
        )
    if cfg_section is None:
        raise ConfigurationError('Missing configuration parameter "pyramid_oereb.cfg.section".')

    try:
        if c2ctemplate_style:
            import c2c.template
            content = c2c.template.get_config(cfg_file)
        else:
            content = pyaml_env.parse_config(cfg_file)
    except IOError as e:
        e.strerror = '{0}{1} \'{2}\', Current working directory is {3}'.format(
            e.strerror, e.args[1], e.filename, os.getcwd())
        raise
    cfg = content.get(cfg_section)
    if cfg is None:
        raise ConfigurationError('YAML file contains no section "{0}"'.format(cfg_section))
    return cfg



# -*- coding: utf-8 -*-
import logging

import datetime
from pyramid_oereb.core.config import Config
from geolink_formatter import XML
from requests.auth import HTTPBasicAuth

from pyramid_oereb.core.records.documents import DocumentRecord
from pyramid_oereb.core.records.office import OfficeRecord
from pyramid_oereb.core.sources import Base


log = logging.getLogger(__name__)


class OEREBlexSource(Base):
    """
    A document source, that creates records for the received documents from OEREBlex for the specified
    geoLink.
    """
    def __init__(self, **kwargs):
        """
        Creates a new OEREBlex document source.

        Keyword Args:
            host (uri): Host URL of OEREBlex (without /api/...).
            version (str): The used geoLink schema version. Default is 1.2.2
            pass_version (bool): True to pass version in URL, false otherwise. Defaults is false.
            language (str): The language of the received data.
            canton (str): Canton code used for the documents.
            mapping (dict of str): Mapping for optional attributes.
            related_decree_as_main (bool): Add related decrees directly to the public law restriction.
            related_notice_as_main (bool): Add related notices directly to the public law restriction.
            proxy (dict of uri): Optional proxy configuration for HTTP and/or HTTPS.
            auth (dict of str): Optional credentials for basic authentication. Requires `username`
                and `password` to be defined.
            validation (bool): Turn XML validation on/off. Default is true.
            url_param_config (list of code and url_param): Optional url parameters to use, per plr code
            code (str): The official code. Regarding to the federal specifications.
            use_prepubs (bool): If true and the law status is not "inForce", the prepubs URL will be
                used. Default is false.

        """
        super(OEREBlexSource, self).__init__()

        # Get keyword arguments
        self._version = kwargs.get('version')
        self._pass_version = kwargs.get('pass_version')
        self._mapping = kwargs.get('mapping')
        self._related_decree_as_main = kwargs.get('related_decree_as_main')
        self._related_notice_as_main = kwargs.get('related_notice_as_main')
        self._proxies = kwargs.get('proxy')
        self._code = kwargs.get('code')
        self._use_prepubs = kwargs.get('use_prepubs')

        log.debug('Use prepubs: {0}'.format(self._use_prepubs))

        # Set default values for missing parameters
        if self._version is None:
            self._version = '1.2.2'
        if self._pass_version is None:
            self._pass_version = False

        auth = kwargs.get('auth')
        if isinstance(auth, dict) and 'username' in auth and 'password' in auth:
            self._auth = HTTPBasicAuth(auth.get('username'), auth.get('password'))
        else:
            self._auth = None

        self._language = str(kwargs.get('language')).lower()
        if not (isinstance(self._language, str) and len(self._language) == 2):
            raise AssertionError('language has to be string of two characters, e.g. "de" or "fr"')

        self._canton = kwargs.get('canton')
        if not (isinstance(self._canton, str) and len(self._canton) == 2):
            raise AssertionError('canton has to be string of two characters, e.g. "BL" or "NE"')

        if kwargs.get('validation') is not None:
            xsd_validation = kwargs.get('validation')
        else:
            xsd_validation = True
        self._parser = XML(host_url=kwargs.get('host'), version=self._version, xsd_validation=xsd_validation)
        if self._parser.host_url is None:
            raise AssertionError('host_url has to be defined')

        self._url_param_config = kwargs.get('url_param_config')
        if self._url_param_config:
            if not (isinstance(self._url_param_config, list)):
                raise AssertionError('url_param_config is of wrong type {}, should be list'
                                     .format(type(self._url_param_config)))
            for list_entry in self._url_param_config:
                if not (isinstance(list_entry, dict)):
                    raise AssertionError('url_param_config list entry is of wrong type {},'
                                         ' should be dictionary'.format(type(list_entry)))

    def read(self, params, geolink_id, law_status, oereblex_params=None):
        """
        Requests the geoLink for the specified ID and returns records for the received documents.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            geolink_id (int): The geoLink ID.
            law_status (pyramid_oereb.core.records.lawstatus.LawStatusRecord): The restriction's law status.
            oereblex_params (string or None): Any additional parameters to pass to Oereblex
        """
        log.debug("read() start for geolink_id {}, oereblex_params {}".format(geolink_id, oereblex_params))

        if self._use_prepubs and law_status.code != 'inForce':
            service = 'prepubs'
        else:
            service = 'geolinks'

        url_base = '{host}/api/{version}{service}/{id}.xml'
        if oereblex_params:
            url_base = url_base + '?' + oereblex_params

        # Request documents
        url = url_base.format(
            host=self._parser.host_url,
            version=self._version + '/' if self._pass_version else '',
            service=service,
            id=geolink_id,
            url_params=oereblex_params
        )

        language = params.language or self._language
        request_params = {
            'locale': language
        }
        log.debug("read() getting documents, url: {}, parser: {}".format(url, self._parser))
        documents = self._parser.from_url(url, request_params, proxies=self._proxies, auth=self._auth)
        log.debug("read() got documents")

        # Convert to records
        self.records = []
        for document in documents:
            self.records.extend(self._get_document_records(document, language))
        log.debug("read() done.")

    def _get_document_records(self, document, language):
        """
        Converts the received documents into records.

        Args:
            document (geolink_formatter.entity.Document): The geoLink document to be returned as document
                record.
            language (str): The language of the returned documents.

        Returns:
            list of pyramid_oereb.core.records.documents.DocumentRecord: The converted record.
        """

        # Cancel if document contains no files
        if len(document.files) == 0:
            log.warning('Document with OEREBlex ID {0} has been skipped because of missing file.'.format(
                document.id
            ))
            return []

        enactment_date = document.enactment_date
        authority = document.authority
        if document.doctype == 'notice':
            # Oereblex notices documents can have no enactment_date while it is require by pyramid_oereb to
            # have one. Add a fake default one that is identifiable and always older than now (01.0.1.1970).
            if enactment_date is None:
                enactment_date = datetime.date(1970, 1, 1)
            # Oereblex notices documents can have no `authority` while it is require by pyramid_oereb to
            # have one. Replace None by '-' in this case.
            if authority is None:
                authority = '-'

        # Cancel if enactment_date is not set
        if enactment_date is None:
            log.warning('Document with OEREBlex ID {0} has been skipped because of missing enactment_date.'
                        .format(document.id))
            return []

        # Check mandatory attributes
        if document.title is None:
            raise AssertionError('Missing title for document #{0}'.format(document.id))
        if authority is None:
            raise AssertionError('Missing authority for document #{0}'.format(document.id))

        # Get document type
        document_type = Config.get_document_type_by_data_code(self._code, document.doctype)

        # Create related office record
        office = OfficeRecord({language: authority}, office_at_web=document.authority_url)

        # Get files
        records = []
        for f in document.files:
            arguments = {
                'document_type': document_type,
                'index': document.index,
                'law_status': Config.get_law_status_by_data_code(self._code, u'inKraft'),
                'title': self._get_document_title(document, f, language),
                'responsible_office': office,
                'published_from': enactment_date,  # TODO: Use "publication_date" instead?
                'published_until': None,  # TODO: Use "abrogation_date"?
                'text_at_web': self._get_multilingual(f.href, language),
                'abbreviation': self._get_multilingual(document.abbreviation, language),
                'official_number': self._get_multilingual(document.number, language),
                'only_in_municipality': None,  # TODO: Use "municipality" from OEREBlex?
                'article_numbers': None
            }
            records.append(DocumentRecord(**arguments))

        return records

    @staticmethod
    def _get_document_title(document, current_file, language):
        """
        Returns the title of the document/file. Extracting this logic allows
        easier customization of the file title.

        Args:
            document (geolink_formatter.entity.Document): The document entity.
            current_file (geolink_formatter.entity.File): The file, which gets annotated with a title.
            language (str): The language of the document title.

        Returns:
            str: Title of document.
        """
        # Assign multilingual values
        return OEREBlexSource._get_multilingual(current_file.title or document.title, language)

    def _get_mapped_value(self, document, key, language=None):
        """
        Return the value of a mapped optional attribute.

        Args:
            document (geolink_formatter.entity.Document): The document entity.
            key (str): The key of the attribute to be mapped.
            language (str or None): Pass language to wrap value in multilingual dictionary.

        Returns:
            str or None: The value of the mapped attribute.

        TODO: Maybe obsolete?
        """
        if self._mapping:
            attribute = self._mapping.get(key)
            if attribute:
                value = getattr(document, attribute)
                if value:
                    return {language: value} if language else value
        return None

    @staticmethod
    def _get_multilingual(value, language):
        """
        Returns the title of the document/file. Extracting this logic allows
        easier customization of the file title.

        Args:
            value (str): The attribute value.
            language (str): The language of the attribute value.

        Returns:
            dict: The multilingual element.
        """

        # Skip for None values
        if value is None:
            return None

        # Assign multilingual values
        return {language: value}



# -*- coding: utf-8 -*-


class AvailabilityRecord(object):

    def __init__(self, fosnr, theme_code, available=False):
        """
        The record to check if data is available for municipality.

        Args:
            fosnr (int): The unique id of the municipality.
            theme_code (str): The theme_code which is affected to be available or not.
            available (bool): Switch if data is available in municipality and theme/sub theme or not.
        """
        self.fosnr = fosnr
        self.theme_code = theme_code
        self.available = available



# -*- coding: utf-8 -*-
from pyramid.path import DottedNameResolver


class AddressReader(object):
    """
    The central reader accessor for addresses. It is directly bound to a so called source which is defined by
    a pythonic dotted string to the class definition of this source. An instance of the passed source will be
    created on instantiation of this reader class by passing through the parameter kwargs.

    The address is here the real combination of a street name, a zip code and a street number.
    """

    def __init__(self, dotted_source_class_path, **params):
        """
        Args:
            dotted_source_class_path (str or pyramid_oereb.lib.sources.address.AddressBaseSource):
                The path to the class which represents the source used by this reader. This
                class must exist and it must implement basic source behaviour of the
                :ref:`api-pyramid_oereb-core-sources-address-addressbasesource`.
            (kwargs): kwargs, which are necessary as configuration parameter for the above by
                dotted name defined class.
        """
        source_class = DottedNameResolver().maybe_resolve(dotted_source_class_path)
        self._source_ = source_class(**params)

    def read(self, params, street_name, zip_code, street_number):
        """
        The read method of this reader. There we invoke the read method of the bound source.

        Note:
            If you subclass this class your implementation needs to offer this method in the same
            signature. Means the parameters must be the same and the return must be a list of
            :ref:`api-pyramid_oereb-core-records-address-addressrecord`. Otherwise the API like way the server
            works would be broken.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            street_name (unicode): The name of the street for the desired address.
            zip_code (int): The postal zipcode for the desired address.
            street_number (str): The house or so called street number of the desired address.

        Returns:
            list of pyramid_oereb.core.records.address.AddressRecord:
                The list of found records filtered by the passed criteria.
        """
        self._source_.read(params, street_name, zip_code, street_number)
        return self._source_.records



# -*- coding: utf-8 -*-
from c2cwsgiutils.sqlalchemylogger.handlers import SQLAlchemyHandler
from pyramid_oereb.core.config import Config
import configparser
import ast
from mako.template import Template
import os
import optparse

SCRIPT_FOLDER = os.path.dirname(os.path.abspath(__file__))


def create_stats_tables():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='creates database and views for sqlalchemylogger'
    )
    parser.add_option(
        '-c', '--config',
        dest='configfile',
        metavar='INI_FILE',
        type='string',
        help='The same config .ini file used to initialize the sqlalchemylogger handler'
    )
    parser.add_option(
        '-s', '--section',
        dest='config_section',
        metavar='SECTION_NAME',
        default='handler_sqlalchemylogger',
        help='section in the ini-file. Default = "handler_sqlalchemylogger"'
    )
    parser.add_option(
        '-a', '--args',
        dest='config_sql_args',
        metavar='SUBSECTION',
        default='args',
        help='subsection in the ini-file. Default = "args"'
    )
    options, _ = parser.parse_args()
    if not options.configfile:
        parser.error('No configfile set')
    _create_views(config_file=options.configfile,
                  config_section=options.config_section,
                  config_sql_args=options.config_sql_args)


def _create_views(config_file,
                  config_section='handler_sqlalchemylogger',
                  config_sql_args='args'):
    config = configparser.ConfigParser(Config.get_db_vars_from_env())
    config.read(config_file)
    schema_name = ast.literal_eval(config[config_section][config_sql_args])[0]['tableargs']['schema']
    tablename = ast.literal_eval(config[config_section][config_sql_args])[0]['tablename']
    fake_handler = SQLAlchemyHandler(ast.literal_eval(config[config_section][config_sql_args])[0])
    fake_handler.create_db()
    create_view_sql = Template(filename='{}/templates/views.sql.mako'.format(SCRIPT_FOLDER))
    fake_handler.session.execute(create_view_sql.render(schema_name=schema_name, tablename=tablename))
    fake_handler.session.commit()



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.office import OfficeRecord


class OfficeBaseSource(Base):
    """
    Base class for office sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.office.OfficeRecord): List of office records.
    """
    _record_class_ = OfficeRecord

    def read(self):
        """
        Every office source has to implement a read method. This method must accept no parameters. Because
        it should deliver all items available.
        If you want adapt to your own source for offices, this is the point where to hook in.
        """
        pass  # pragma: no cover



from sqlalchemy import String, Integer
from sqlalchemy.ext.declarative import declarative_base

from pyramid_oereb.contrib.data_sources.standard.models import (
    get_office,
    get_document,
    get_view_service,
    get_legend_entry,
    get_public_law_restriction,
    get_geometry,
    get_public_law_restriction_document
)


class Models(object):

    def __init__(self, office, document, view_service,
                 legend_entry, public_law_restriction, geometry,
                 public_law_restriction_document, base, db_connection, schema_name):

        self.Office = office
        self.Document = document
        self.ViewService = view_service
        self.LegendEntry = legend_entry
        self.PublicLawRestriction = public_law_restriction
        self.Geometry = geometry
        self.PublicLawRestrictionDocument = public_law_restriction_document
        self.Base = base
        self.db_connection = db_connection
        self.schema_name = schema_name


def model_factory(schema_name, pk_type, geometry_type, srid, db_connection):
    """
    Factory to produce a set of standard models.

    Args:
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        db_connection (str): the db connection string

    Returns:
        Models: the produced set of standard models
    """
    Base = declarative_base()

    Office = get_office(Base, schema_name, pk_type)
    Document = get_document(Base, schema_name, pk_type, Office)
    ViewService = get_view_service(Base, schema_name, pk_type)
    LegendEntry = get_legend_entry(Base, schema_name, pk_type, ViewService)
    PublicLawRestriction = get_public_law_restriction(Base, schema_name, pk_type, Office, ViewService,
                                                      LegendEntry)
    Geometry = get_geometry(Base, schema_name, pk_type, geometry_type, srid, PublicLawRestriction)
    PublicLawRestrictionDocument = get_public_law_restriction_document(Base, schema_name, pk_type,
                                                                       PublicLawRestriction, Document)

    return Models(
        Office, Document, ViewService,
        LegendEntry, PublicLawRestriction, Geometry, PublicLawRestrictionDocument,
        Base, db_connection, schema_name
    )


def model_factory_string_pk(schema_name, geometry_type, srid, db_connection):
    """
    Args:
        schema_name (str): The name of the database schema where this models belong to.
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        db_connection (str): the db connection string

    Returns:
        Models: the produced set of standard models
    """
    return model_factory(schema_name, String, geometry_type, srid, db_connection)


def model_factory_integer_pk(schema_name, geometry_type, srid, db_connection):
    """
    Args:
        schema_name (str): The name of the database schema where this models belong to.
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        db_connection (str): the db connection string

    Returns:
        Models: the produced set of standard models
    """
    return model_factory(schema_name, Integer, geometry_type, srid, db_connection)



from sqlalchemy import Column, ForeignKey
from sqlalchemy import String, Integer, Date
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from pyramid_oereb.contrib.data_sources.standard.models import (
    get_office,
    get_view_service,
    get_legend_entry,
    get_geometry
)


class Models(object):

    def __init__(self, office, view_service,
                 legend_entry, public_law_restriction, geometry, base,
                 db_connection, schema_name):

        self.Office = office
        self.ViewService = view_service
        self.LegendEntry = legend_entry
        self.PublicLawRestriction = public_law_restriction
        self.Geometry = geometry
        self.Base = base
        self.db_connection = db_connection
        self.schema_name = schema_name


def model_factory(schema_name, pk_type, geometry_type, srid, db_connection):
    """
    Factory to produce a set of standard models.

    Args:
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
    """
    Base = declarative_base()

    Office = get_office(Base, schema_name, pk_type)
    ViewService = get_view_service(Base, schema_name, pk_type)
    LegendEntry = get_legend_entry(Base, schema_name, pk_type, ViewService)

    class PublicLawRestriction(Base):
        """
        The container where you can fill in all your public law restrictions to the topic.

        Attributes:
            id (str): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            law_status (str): The status switch if the document is legally approved or not.
            published_from (datetime.date): The date when the document should be available for
                publishing on extracts. This  directly affects the behaviour of extract
                generation.
            published_until (datetime.date): The date starting from which the document should not be
                published anymore on extracts. This directly affects the behaviour of extract generation.
            geolink (int): The OEREBlex GEO-Link ID to query the documents.
            view_service_id (str): The foreign key to the view service this public law restriction is
                related to.
            view_service (ViewService):
                The dedicated relation to the view service instance from database.
            office_id (str): The foreign key to the office which is responsible to this public law
                restriction.
            responsible_office (Office):
                The dedicated relation to the office instance from database.
            legend_entry_id (str): The foreign key to the legend entry this public law restriction is
                related to.
            legend_entry (pyramid_oereb.standard.models.airports_building_lines.LegendEntry):
                The dedicated relation to the legend entry instance from database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'public_law_restriction'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        law_status = Column(String, nullable=False)
        published_from = Column(Date, nullable=False)
        published_until = Column(Date, nullable=True)
        geolink = Column(Integer, nullable=False)
        view_service_id = Column(
            ForeignKey(ViewService.id),
            nullable=False
        )
        view_service = relationship(
            ViewService,
            backref='public_law_restrictions'
        )
        office_id = Column(
            ForeignKey(Office.id),
            nullable=False
        )
        responsible_office = relationship(Office)
        legend_entry_id = Column(
            ForeignKey(LegendEntry.id),
            nullable=False
        )
        legend_entry = relationship('LegendEntry', backref='public_law_restrictions')

    Geometry = get_geometry(Base, schema_name, pk_type, geometry_type, srid, PublicLawRestriction)

    return Models(
        Office, ViewService,
        LegendEntry, PublicLawRestriction, Geometry, Base,
        db_connection, schema_name
    )


def model_factory_string_pk(schema_name, geometry_type, srid, db_connection):
    """
    Args:
        schema_name (str): The name of the database schema where this models belong to.
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        db_connection (str): the db connection string

    Returns:
        Models: the produced set of standard models
    """
    return model_factory(schema_name, String, geometry_type, srid, db_connection)


def model_factory_integer_pk(schema_name, geometry_type, srid, db_connection):
    """
    Args:
        schema_name (str): The name of the database schema where this models belong to.
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        db_connection (str): the db connection string

    Returns:
        Models: the produced set of standard models
    """
    return model_factory(schema_name, Integer, geometry_type, srid, db_connection)



# -*- coding: utf-8 -*-

import logging
import warnings
from datetime import datetime


log = logging.getLogger(__name__)


class EmptyPlrRecord(object):
    """
    Attributes:
        theme (pyramid_oereb.lib.records.theme.ThemeRecord): The theme to which the PLR belongs to.
        has_data (bool): True if the topic contains data.
    """

    def __init__(self, theme, has_data=True):
        """
        Record for empty topics.

        Args:
            theme (pyramid_oereb.lib.records.theme.ThemeRecord): The theme to which the PLR belongs to.
            has_data (bool): True if the topic contains data.
        """
        self.theme = theme
        self.has_data = has_data


class PlrRecord(EmptyPlrRecord):
    """
    Public law restriction record.
    Attributes:
        theme (pyramid_oereb.lib.records.theme.ThemeRecord): The theme to which the PLR belongs to.
        legend_entry (pyramid_oereb.lib.records.view_service.LegendEntryRecord): The PLR record's
            corresponding legend record.
        law_status (pyramid_oereb.lib.records.law_status.LawStatusRecord): The law status of this record.
        published_from (datetime.date): Date from/since when the PLR record is published.
        published_until (datetime.date): Date from when the PLR record is not published anymore.
        responsible_office (pyramid_oereb.lib.records.office.OfficeRecord): Office which is responsible
            for this PLR.
        symbol (pyramid_oereb.lib.records.image.ImageRecord): Symbol of the restriction defined for the
            legend entry
        view_service (pyramid_oereb.lib.records.view_service.ViewServiceRecord): The view service instance
            associated with this record.
        geometries (list of pyramid_oereb.lib.records.geometry.GeometryRecord): List of geometry records
            associated with this record.
        sub_theme (pyramid_oereb.lib.records.theme.ThemeRecord or None): Optional subtopic.
        type_code (unicode): The PLR record's type code (also used by view service).
        type_code_list (unicode): URL to the PLR's list of type codes.
        documents (list of pyramid_oereb.core.records.documents.DocumentBaseRecord): List of documents
            associated with this record.
        info (dict or None): The information read from the config.
        min_length (float): The threshold for area calculation.
        min_area (float): The threshold for area calculation.
        length_unit (unicode): The threshold for area calculation.
        area_unit (unicode): The threshold for area calculation.
        view_service_id (int): The id to the connected view service. This is very important to be able to
        solve bug https://github.com/openoereb/pyramid_oereb/issues/521
    """

    def __init__(self, theme, legend_entry, law_status, published_from, published_until, responsible_office,
                 symbol, view_service, geometries, sub_theme=None, type_code=None,
                 type_code_list=None, documents=None, info=None, min_length=0.0,
                 min_area=0.0, length_unit=u'm', area_unit=u'm2', view_service_id=None):
        """
        Args:
            theme (pyramid_oereb.lib.records.theme.ThemeRecord): The theme to which the PLR belongs to.
            legend_entry (pyramid_oereb.lib.records.view_service.LegendEntryRecord): The PLR record's
                corresponding legend record.
            law_status (pyramid_oereb.lib.records.law_status.LawStatusRecord): The law status of this record.
            published_from (datetime.date): Date from/since when the PLR record is published.
            published_until (datetime.date): Date from when the PLR record is not published anymore.
            responsible_office (pyramid_oereb.lib.records.office.OfficeRecord): Office which is responsible
                for this PLR.
            symbol (pyramid_oereb.lib.records.image.ImageRecord): Symbol of the restriction defined for the
                legend entry
            view_service (pyramid_oereb.lib.records.view_service.ViewServiceRecord): The view service instance
                associated with this record.
            geometries (list of pyramid_oereb.lib.records.geometry.GeometryRecord): List of geometry records
                associated with this record.
            sub_theme (pyramid_oereb.lib.records.theme.ThemeRecord or None): Optional subtopic.
            type_code (unicode): The PLR record's type code (also used by view service).
            type_code_list (unicode): URL to the PLR's list of type codes.
            documents (list of pyramid_oereb.core.records.documents.DocumentBaseRecord): List of documents
                associated with this record.
            info (dict or None): The information read from the config.
            min_length (float): The threshold for area calculation.
            min_area (float): The threshold for area calculation.
            length_unit (unicode): The threshold for area calculation.
            area_unit (unicode): The threshold for area calculation.
            view_service_id (int): The id to the connected view service. This is very important to be able to
            solve bug https://github.com/openoereb/pyramid_oereb/issues/521
        """
        super(PlrRecord, self).__init__(theme)

        if not isinstance(legend_entry.legend_text, dict):
            warnings.warn('Type of "legend_text" should be "dict"')

        assert isinstance(geometries, list)
        assert len(geometries) > 0

        self.legend_entry = legend_entry
        self.law_status = law_status
        self.published_from = published_from
        self.published_until = published_until
        self.responsible_office = responsible_office
        self.type_code = type_code
        self.type_code_list = type_code_list
        self.view_service = view_service
        if documents is None:
            self.documents = []
        else:
            self.documents = documents
        self.geometries = geometries
        self.sub_theme = sub_theme
        self.info = info
        self.has_data = True
        self.min_length = min_length
        self.min_area = min_area
        self.area_unit = area_unit
        self.length_unit = length_unit
        self._area_share = None
        self._part_in_percent = None
        self._length_share = None
        self._nr_of_points = None
        self.symbol = symbol
        self.view_service_id = view_service_id

    @property
    def legend_text(self):
        """
        Returns:
            dict of unicode: the legend text
        """
        return self.legend_entry.legend_text

    @property
    def published(self):
        """
        Retruns:
            bool: True if PLR is published.
        """
        if self.published_until is None:
            return self.published_from <= datetime.now().date()
        else:
            return self.published_from <= datetime.now().date() \
                   and self.published_until >= datetime.now().date()

    def _sum_length(self):
        """
        Returns:
            float: The summed length.
        """
        lengths_to_sum = []
        for geometry in self.geometries:
            if geometry.length_share:
                lengths_to_sum.append(geometry.length_share)
        return sum(lengths_to_sum) if len(lengths_to_sum) > 0 else None

    def _sum_area(self):
        """
        Returns:
            float: The summed area.
        """
        areas_to_sum = []
        for geometry in self.geometries:
            if geometry.area_share:
                areas_to_sum.append(geometry.area_share)
        return sum(areas_to_sum) if len(areas_to_sum) > 0 else None

    def _sum_points(self):
        """
        Returns:
            int: The summed number of points of these geometry records.
        """
        points_to_sum = 0
        for geometry in self.geometries:
            if geometry.nr_of_points:
                points_to_sum += geometry.nr_of_points
        return points_to_sum

    @property
    def area_share(self):
        """
        Retruns:
            float or None: the summed area of all related geometry records of this PLR."""
        return self._area_share

    @property
    def part_in_percent(self):
        """
        Retruns:
            decimal or None: (decimal): Part of the property area touched by the restriction in percent."""
        return self._part_in_percent

    @part_in_percent.setter
    def part_in_percent(self, value):
        """
        Args:
            value (decimal): Part of the property area touched by the restriction in percent.
        """
        self._part_in_percent = value

    @property
    def length_share(self):
        """
        Returns:
            float or None: the summed length of all related geometry records of this PLR."""
        return self._length_share

    @property
    def nr_of_points(self):
        """
        Returns:
            float or None: the number of points of all related geometry records of this PLR."""
        return self._nr_of_points

    def calculate(self, real_estate, geometry_types):
        """
        Entry method for calculation. It checks if the geometry type of this instance is a geometry
        collection which has to be unpacked first in case of collection.

        Args:
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real estate record.
            geometry_types (dict): The allowed geometry types for the to match the simple
            feature types point, line, polygon

        Returns:
            bool: True if intersection fits the limits.
        """
        tested_geometries = []
        inside = False
        for geometry in self.geometries:
            if geometry.published and geometry.calculate(
                    real_estate,
                    self.min_length, self.min_area,
                    self.length_unit, self.area_unit,
                    geometry_types
            ):
                tested_geometries.append(geometry)
                inside = True
        self.geometries = tested_geometries

        # Points
        nr_of_points = self._sum_points()
        self._nr_of_points = nr_of_points if nr_of_points else None
        # Lines
        length_share = self._sum_length()
        if length_share is None:
            self._length_share = None
        else:
            self._length_share = int(round(length_share, 0))
        # Areas
        area_share = self._sum_area()
        if area_share is None:
            self._area_share = None
            self._part_in_percent = None
        else:
            self._area_share = int(round(area_share, 0))
            self._part_in_percent = round(
                ((float(self._area_share) / float(real_estate.land_registry_area)) * 100),
                1
            )
        return inside

    def __str__(self):
        legend_text = dict()
        for key in self.legend_text:
            if self.legend_text[key] is not None:
                legend_text[key] = self.legend_text[key].encode('utf-8')
        tpl = '<{} -- type_code: {} theme: {} legend_text: {} (further attributes not shown)>'
        return tpl.format(self.__class__.__name__, self.type_code, self.theme, legend_text)



# -*- coding: utf-8 -*-
from pyramid.path import DottedNameResolver


class DocumentReader(object):
    """
    The central reader for the document definitions. It is directly bound to a so called source
    which is defined by a pythonic dotted string to the class definition of this source.
    An instance of the passed source will be created on instantiation of this reader class by passing through
    the parameter kwargs.
    """

    def __init__(self, dotted_source_class_path, **params):
        """
        Args:
            dotted_source_class_path
                (str or pyramid_oereb.core.sources.document.DocumentBaseSource): The path to
                the class which represents the source used by this reader. This class must
                exist and it must implement basic source behaviour of the
                :ref:`api-pyramid_oereb-core-sources-document-documentbasesource`.
            (kwargs): kwargs, which are necessary as configuration parameter for the above by
                dotted name defined class.
        """
        source_class = DottedNameResolver().maybe_resolve(dotted_source_class_path)
        self._source_ = source_class(**params)

    def read(self, office_records):
        """
        The central read accessor method to get all desired records from configured source.

        Note:
            If you subclass this class your implementation needs to offer this method in the same
            signature. Means the parameters must be the same and the return must be a list of
            :ref:`api-pyramid_oereb-core-records-documents-documentrecord`. Otherwise the API like way
            the server works would be broken.

        Returns:
            list of pyramid_oereb.core.records.documents.DocumentRecord:
                The list of found records. Since these are not filtered by any criteria the list simply
                contains all records delivered by the source.
        """
        self._source_.read(office_records)
        return self._source_.records



# -*- coding: utf-8 -*-
from geoalchemy2.elements import _SpatialElement

from pyramid_oereb.core.sources import BaseDatabaseSource
from geoalchemy2.shape import to_shape

from pyramid_oereb.core.sources.real_estate import RealEstateBaseSource


class DatabaseSource(BaseDatabaseSource, RealEstateBaseSource):

    def read(self, params, nb_ident=None, number=None, egrid=None, geometry=None):
        """
        Central method to read all plrs (geometry input) or explicitly one plr (nb_ident+number/egrid input).

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            nb_ident (int or None): The identification number of the desired real estate. This
                parameter is directly related to the number parameter and both must be set!
                Combination will deliver only one result or crashes.
            number (str or None): The number of parcel or also known real estate. This parameter
                is directly related to the nb_ident parameter and both must be set!
                Combination will deliver only one result or crashes.
            (str or None): The unique identifier of the desired real estate. This will deliver
                only one result or crashes.
            geometry (str or None): A geometry as WKT string which is used to obtain intersected real
                estates. This may deliver several results.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            query = session.query(self._model_)
            if nb_ident and number:
                results = query.filter(self._model_.number == number, self._model_.identdn == nb_ident).all()
            elif egrid:
                results = query.filter(self._model_.egrid == egrid).all()
            elif geometry:
                results = query.filter(self._model_.limit.ST_Intersects(geometry)).all()
            else:
                raise AttributeError('Necessary parameter were missing.')

            self.records = list()
            for result in results:
                self.records.append(self._record_class_(
                    result.type,
                    result.canton,
                    result.municipality,
                    result.fosnr,
                    result.land_registry_area,
                    to_shape(result.limit) if isinstance(result.limit, _SpatialElement) else None,
                    metadata_of_geographical_base_data=result.metadata_of_geographical_base_data,
                    number=result.number,
                    identdn=result.identdn,
                    egrid=result.egrid,
                    subunit_of_land_register=result.subunit_of_land_register,
                    subunit_of_land_register_designation=result.subunit_of_land_register_designation
                ))

        finally:
            session.close()



# -*- coding: utf-8 -*-

from geoalchemy2.elements import _SpatialElement

from pyramid_oereb.core.sources import BaseDatabaseSource
from geoalchemy2.shape import to_shape

from pyramid_oereb.core.sources.municipality import MunicipalityBaseSource


class DatabaseSource(BaseDatabaseSource, MunicipalityBaseSource):

    def read(self):
        """
        The read method to access the standard database structure. It uses SQL-Alchemy for querying. It does
        not accept any parameters nor it applies any filter on the database query. It simply loads all
        content from the configured model.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            self.records = list()
            results = session.query(self._model_).all()
            for result in results:
                self.records.append(self._record_class_(
                    result.fosnr,
                    result.name,
                    result.published,
                    geom=to_shape(result.geom).wkt if isinstance(
                        result.geom, _SpatialElement) else None,
                ))
        finally:
            session.close()



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.view_service import LegendEntryRecord


class LegendBaseSource(Base):
    """
    Base class for exclusion of liability sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.view_service.LegendEntryRecord): List of legend entry
            records.
    """
    _record_class_ = LegendEntryRecord

    def read(self, params, **kwargs):
        """
        Every legend entry source has to implement a read method. If you want adapt to your own source for
        legend entries, this is the point where to hook in.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            (kwargs): Arbitrary keyword arguments.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import logging
import textwrap

log = logging.getLogger(__name__)


class TocPages():

    def __init__(self, extract):
        # variables taken from template toc.jrxml
        self.disposable_height = 842 - 70  # A4 size - (footer + header)
        self.d1_height = 77
        self.d2_height = 29
        self.d3_height = 61
        self.d4_height = 44
        self.d5_height = 15
        self.d6_height = 90
        self.d6_right_height = 23
        self.d6_right_width = 233
        self.d6_stuff_y_location = 39
        self.d6_left_height = 0  # FIXME: compute this
        self.title_size = 62
        self.toc_title_height = 15 + 62 + 12  # height + location + item starting position
        self.toc_item_height = 20
        self.not_concerned_themes_title_height = 15 + 26  # height + location
        self.not_concerned_themes_item_height = 12
        self.theme_without_data_title_height = 12
        self.theme_without_data_item_height = 12
        self.extract = extract
        self.total_length = self.compute_total_lenght()

    def compute_d1(self):
        return self.d1_height

    def compute_d2(self):
        x = len(self.extract['ConcernedTheme'] * self.toc_item_height)
        if x > self.d2_height:
            return x
        else:
            return self.d2_height

    def compute_d3(self):
        x = self.not_concerned_themes_title_height + len(self.extract['NotConcernedTheme'] * self.not_concerned_themes_item_height)  # noqa
        if x > self.d3_height:
            return x
        else:
            return self.d3_height

    def compute_d4(self):
        return self.d4_height

    def compute_d5(self):
        x = len(self.extract['ThemeWithoutData'] * self.theme_without_data_item_height)
        if x > self.d5_height:
            return x
        else:
            return self.d5_height

    def compute_d6_left(self):
        # TODO: compute height when QR-Code is integrated
        content_min_size = 10 + 10 + 5 + 10 + 10  # spacing between paragraphs
        total_size = 39
        paragraph_space = 11
        for i in self.extract.get('GeneralInformation', []):
            total_size += paragraph_space
            total_size += self.compute_length_of_wrapped_text(i[0]['Text'],
                                                              78,
                                                              10)
        log.debug('d6 left total_size : {}'.format(total_size))
        if total_size > content_min_size:
            return total_size
        else:
            return content_min_size

    @staticmethod
    def compute_length_of_wrapped_text(text, nb_char, font_size):
        t = textwrap.wrap(text, nb_char)
        return len(t) * font_size

    def compute_d6_right(self):
        # variables taken from template disclaimer.jrxml
        space_above = 4
        space_title_content = 2
        content_min_size = 23
        total_size = 0
        for i in self.extract.get('Disclaimer', []):
            total_size += space_above
            total_size += self.compute_length_of_wrapped_text(i['Title'][0]['Text'],
                                                              65,
                                                              14)
            total_size += space_title_content
            total_size += self.compute_length_of_wrapped_text(i['Content'][0]['Text'],
                                                              78,
                                                              10)
        log.debug('d6 ritght total_size : {}'.format(total_size))
        if total_size > content_min_size:
            return total_size
        else:
            return content_min_size

    def compute_d6(self):
        x = max(self.compute_d6_left(), self.compute_d6_right()) + self.d6_stuff_y_location
        if x > self.d6_height:
            return x
        else:
            return self.d6_height

    def compute_total_lenght(self):
        x = self.compute_d1() + \
            self.compute_d2() + \
            self.compute_d3() + \
            self.compute_d4() + \
            self.compute_d5() + \
            self.compute_d6()
        log.debug('TOC total page length : {}'.format(x))
        return x

    def getNbPages(self):
        return -(-self.total_length // self.disposable_height)  # ceil number of pages needed



# -*- coding: utf-8 -*-


class MapLayeringRecord(object):
    """
    Represents a map layering entry with the view service, layer index and layer opacity.

    Attributes:
        view_service (dict): View service.
        layer_index (int): Index for sorting the layering of the view services within a theme.
        layer_opacity (float): Opacity of a view service.
    """
    def __init__(self, view_service, layer_index, layer_opacity):
        """
        Args:
            view_service (dict): View service.
            layer_index (int): Index for sorting the layering of the view services within a theme.
            layer_opacity (float): Opacity of a view service.
        """
        self.view_service = view_service
        self.layer_index = layer_index
        self.layer_opacity = layer_opacity



import logging
import json

LOG = logging.getLogger('JSON')


def log_response(wrapped):
    def wrapper(context, request):
        response = wrapped(context, request)
        ret_dict = _serialize_response(response)
        ret_dict.update(_serialize_request(request))
        LOG.info(json.dumps(ret_dict))
        return response
    return wrapper


def _serialize_response(response):
    x = {}
    x['status_code'] = response.status_code
    x['headers'] = dict(response.headers)
    x['extras'] = response.extras if hasattr(response, 'extras') else None
    return {'response': x}


def _serialize_request(request):
    x = {}
    x['headers'] = dict(request.headers)
    x['traversed'] = str(request.traversed)
    x['parameters'] = dict(request.GET)
    x['path'] = str(request.path)
    x['view_name'] = str(request.view_name)
    return {'request': x}


class OerebStats(dict):
    """
    class OerebStats(dict)
    this class is used to provide a serializable object that can be
    used to provide insight of application usage in the logs.
    """
    def __init__(self,
                 service=None,
                 output_format=None,
                 params=None):
        super(OerebStats, self).__init__(service=service,
                                         output_format=output_format,
                                         params=params)
        self.itemlist = super(OerebStats, self).keys()

    def __setitem__(self, key, value):
        super(OerebStats, self).__setitem__(key, value)

    def __iter__(self):
        return iter(self.itemlist)

    def keys(self):
        return self.itemlist

    def values(self):
        return [self[key] for key in self]

    def itervalues(self):
        return (self[key] for key in self)



# -*- coding: utf-8 -*-
from pyramid_oereb.core.sources import Base
from pyramid_oereb.core.records.logo import LogoRecord


class LogoBaseSource(Base):
    """
    Base class for logo images source.

    Attributes:
        records (list of pyramid_oereb.lib.records.logo.LogoRecord): List of logos
        type records.
    """
    _record_class_ = LogoRecord

    def read(self):
        """
        Every logo source has to implement a read method. This method must accept no parameters.
        Because it should deliver all items available.
        If you want adapt to your own source for logo images, this is the point where to hook in.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import optparse

from mako.template import Template
from pyramid.path import AssetResolver


def _create_yaml_config_(
        name='pyramid_oereb.yml',
        database='postgresql://postgres:password@localhost/pyramid_oereb',
        print_backend='MapFishPrint',
        print_url='http://oereb-print:8080/print/oereb'):
    """
    Creates the specified YAML file using a template. This YAML file contains the standard
    configuration to run a oereb server out of the box.

    Args:
        (str): The name of the new file. Default
        database (str): The database connection string.Default:
            'postgresql://postgres:password@localhost/pyramid_oereb'
    """

    # Create pyramid_oereb.yml from template
    template = Template(
        filename=AssetResolver('dev').resolve('config/pyramid_oereb.yml.mako').abspath(),
        input_encoding='utf-8',
        output_encoding='utf-8'
    )
    config = template.render(
        sqlalchemy_url=database,
        print_backend=print_backend,
        print_url=print_url
    )
    pyramid_oereb_yml = open(name, 'wb+')
    pyramid_oereb_yml.write(config)
    pyramid_oereb_yml.close()


def create_yaml():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='Create all content for the standard database'
    )
    parser.add_option(
        '-n', '--name',
        dest='name',
        metavar='YAML',
        type='string',
        default='pyramid_oereb_standard.yml',
        help='The name for the new configuration yaml file (default is: pyramid_oereb_standard.yml).'
    )
    parser.add_option(
        '-d', '--database',
        dest='database',
        metavar='DATABASE',
        type='string',
        default='postgresql://postgres:password@oereb-db:5432/pyramid_oereb',
        help='The database connection string (default is: '
             'postgresql://postgres:password@oereb-db:5432/pyramid_oereb).'
    )
    parser.add_option(
        '-p', '--print_backend',
        dest='print_backend',
        metavar='PRINT_BACKEND',
        type='string',
        default='MapFishPrint',
        help='The print backend (for PDF generation) to use (default is: MapFishPrint)'
    )
    parser.add_option(
        '-u', '--print_url',
        dest='print_url',
        metavar='PRINT_URL',
        type='string',
        default='http://oereb-print:8080/print/oereb',
        help='The URL of the print server'
    )
    options, args = parser.parse_args()
    _create_yaml_config_(
        name=options.name,
        database=options.database,
        print_backend=options.print_backend,
        print_url=options.print_url
    )



# -*- coding: utf-8 -*-
import logging
import binascii

from pyramid.httpexceptions import HTTPNotFound, HTTPServerError

from pyramid_oereb.contrib.data_sources.standard.sources.plr import StandardThemeConfigParser
from pyramid_oereb import database_adapter
from pyramid_oereb.core import b64
from pyramid_oereb.core.records.image import ImageRecord


log = logging.getLogger(__name__)


def get_symbol(params, theme_config):
    """
    Returns the symbol for the requested theme and type code from database. It queries the model
    for the legend entry pyramid_oereb.contrib.data_sources.standard.models.get_legend_entry

    Args:
        params (dict): The URL parameters which were handed over via request.
        theme_config (dict): The configuration of the theme how it was set up in the YAML.

    Returns:
        bytearray, str: The image content and the mimetype of image.
    Raises:
        HTTPNotFound
        HTTPServerError
    """
    try:
        identifier = params['identifier']
    except KeyError:
        log.error(
            f'No useful params for HookMethod was delivered. '
            f'Expected "identifier" in params. Sent params are:  {params}'
        )
        raise HTTPServerError

    config_parser = StandardThemeConfigParser(**theme_config)
    session = database_adapter.get_session(config_parser.db_connection)

    log_string = 'identifier: {}'.format(
        identifier
    )

    try:
        models = config_parser.get_models()
        model = models.LegendEntry
        legend_entry = session.query(model).filter(
            model.t_id == identifier
        ).one()
        if legend_entry:
            symbol = getattr(legend_entry, 'symbol', None)
            if symbol:
                if isinstance(symbol, str):
                    body = b64.decode(symbol)
                elif isinstance(symbol, bytes):
                    body = b64.decode(binascii.b2a_base64(symbol).decode('ascii'))
                else:
                    log.error(f'Symbol was not str nor bytes type but {type(symbol)}. this is not supported.')
                    raise HTTPServerError
                content_type = ImageRecord.get_mimetype(bytearray(body))
                return body, content_type
            else:
                log.error(f'No symbol definition is available for legend entry {log_string}')
                raise HTTPServerError
        else:
            log.error(f'No legend entry was found in data for {log_string}')
            raise HTTPNotFound
    finally:
        session.close()
